#!/usr/bin/env python
# coding: utf-8

"""

    Salsa20: a Fast Streaming Cipher
    --------------------------------

    Salsa20 is a fast stream cipher written by Daniel Bernstein 
    that basically uses a hash function and XOR making for fast 
    encryption.  (Decryption uses the same function.)  Salsa20 
    is simple and quick.  
    
    Some Salsa20 parameter values...
        design strength    128 bits
        key length         128 or 256 bits, exactly
        IV, aka nonce      64 bits, always
        chunk size         must be in multiples of 64 bytes
    
    Salsa20 has two reduced versions, 8 and 12 rounds each.
    
    One benchmark (10 MB):
        1.5GHz PPC G4     102/97/89 MB/sec for 8/12/20 rounds
        AMD Athlon 2500+   77/67/53 MB/sec for 8/12/20 rounds
          (no I/O and before Python GC kicks in)
    
    Salsa20 is a Phase 3 finalist in the EU eSTREAM competition 
    and appears to be one of the fastest ciphers.  It is well 
    documented so I will not attempt any injustice here.  Please 
    see "References" below.
    
    ...and Salsa20 is "free for any use".  
    
    
    pySalsa20: a Python wrapper for Salsa20
    ---------------------------------------

    pySalsa20.py is a simple ctypes Python wrapper.  Salsa20 is 
    as it's name implies, 20 rounds, but there are two reduced 
    versions, 8 and 12 rounds each.  Because the APIs are 
    identical, pySalsa20 is capable of wrapping all three 
    versions (number of rounds hardcoded), including a special 
    version that allows you to set the number of rounds with a 
    set_rounds() function.  Compile the version of your choice 
    as a shared library (not as a Python extension), name and 
    install it as libsalsa20.so.
    
    Sample usage:
        from pySalsa20 import Salsa20
        s20 = Salsa20(key, IV)
        dataout = s20.encryptBytes(datain)   # same for decrypt
    
    This is EXPERIMENTAL software and intended for educational 
    purposes only.  To make experimentation less cumbersome, 
    pySalsa20 is also free for any use.      
    
    THIS PROGRAM IS PROVIDED WITHOUT WARRANTY OR GUARANTEE OF
    ANY KIND.  USE AT YOUR OWN RISK.  
    
    Enjoy,
      
    Larry Bugbee
    bugbee@seanet.com
    April 2007

    
    References:
    -----------
      http://en.wikipedia.org/wiki/Salsa20
      http://en.wikipedia.org/wiki/Daniel_Bernstein
      http://cr.yp.to/djb.html
      http://www.ecrypt.eu.org/stream/salsa20p3.html
      http://www.ecrypt.eu.org/stream/p3ciphers/salsa20/salsa20_p3source.zip
     
    Prerequisites:
    -------------
      - Python 2.5 (or 2.4 with ctypes added)
      - gcc 4.x (gcc 3 should work)
    
    Installation:
    -------------
      Compile provided .c source (set_rounds() mod included) or 
      download salsa20_p3source.zip (see refs) and compile (you 
      may want to apply the changes for set_rounds(), or not ;-)
        PPC MacOSX:
          export MACOSX_DEPLOYMENT_TARGET=10.4
          gcc -fPIC -maltivec -O3 -dynamic -bundle          \
                    -undefined dynamic_lookup               \
                    -o libsalsa20.so  ecrypt-altivec-ppc.c
        Linux:
          gcc -fPIC -O3 -shared -o libsalsa20.so  ecrypt-linux.c
    
      Install libsalsa20.so:
          sudo cp libsalsa20.so /usr/local/lib   # Mandriva wanted /usr/lib
      
      Quick test:
        python pySalsa20.py
    
      Install pySalsa20.py:
        sudo python setup.py install
        
"""

from ctypes import *
import sys

_version = '3'

#--------------------------------------------------------------------------

class Salsa20(object):
    def __init__(self, key=None, IV=None, rounds=None):
        self._lastChunk64 = True
        self._IVbitlen = 64             # must be 64 bits
        self._ctx = c_buffer(64)        # 16 32-bit words
        if key:
            self.setKey(key)
        if IV:
            self.setIV(IV)
        if rounds:
            self.setRounds(rounds)

    def setKey(self, key):
        numbits = len(key)*8
        assert numbits in (128, 256), 'key not 128 or 256 bits'
        libSalsa20.ECRYPT_keysetup(byref(self._ctx), 
                                   key, numbits, self._IVbitlen)
        
    def setIV(self, IV):
        assert len(IV)*8 == 64, 'nonce (IV) not 64 bits'
        libSalsa20.ECRYPT_ivsetup(byref(self._ctx), IV)
    
    setNonce = setIV            # support an alternate name

    def setRounds(self, rounds):
        assert rounds in [8, 12, 20], 'rounds must be 8, 12, 20'
        try:
            libSalsa20.set_rounds(rounds)
        except:
            msg  = '*** your libsalsa20 does not support '  \
                 + 'set_rounds() function; the number of '  \
                 + 'rounds is unknown; probably 20 but '    \
                 + 'you best check ***'
            print msg
#            raise Exception(msg)

    def encryptBytes(self, data):
        assert type(data) == type(''), 'data must be byte string'
        assert self._lastChunk64, 'previous chunk not multiple of 64 bytes'
        lendata = len(data)
        munged = c_buffer(lendata)
        libSalsa20.ECRYPT_encrypt_bytes(byref(self._ctx), 
                                        c_buffer(data), 
                                        byref(munged), 
                                        lendata)
        self._lastChunk64 = not lendata % 64
        return munged.raw[:lendata]
    
    decryptBytes = encryptBytes # encrypt and decrypt use same function


# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# load the shared library into ctypes

def loadLib(name):
    # prefix might need to change for some platforms       ???
    prefix = ''
    # get the correct library suffix
    libsuffixes = {'darwin': '.so',     # .dylib           ???
                   'linux':  '.so', 
                   'linux2': '.so', 
                   'win32':  '.dll'}    # .lib             ???
    try:
        libsuffix = libsuffixes[sys.platform]
    except:
        raise Exception('library suffix for "%s" is what?' % 
                                                  sys.platform)
    libname = prefix+'lib'+name+libsuffix
    return CDLL(libname)                # load and return the library


libSalsa20 = loadLib('salsa20')


#--------------------------------------------------------------------------
# utilities

def savetofile(filename, content):
    "write content to a file [as binary]"
    f = open(filename, 'wb')
    f.write(content)
    f.close()
    
def loadfmfile(filename):
    "get [binary] content from a file"
    f = open(filename, 'rb')
    content = f.read()
    f.close()
    return content

def bytestring(hex):
    "remove whitespace and convert hex string to a byte string"
    return hex.replace(' ', '').replace('\n', '').decode('hex')

#--------------------------------------------------------------------------
# a quick self-test.  Invoke with:  python pySalsa20.py

def test():
    print '  pySalsa20 version %s' % _version

    if 1:
        rounds  = 8                     # may be 8, 12, or 20
    
        if 0:
            message = loadfmfile('testdata.txt')
        else:
            message = 'Kilroy was here!  ...there, and everywhere.'
        key     = 'myKey67890123456'    # 16 or 32 bytes, exactly
        nonce   = 'aNonce'              # do better in real life
        IV      = (nonce+'*'*8)[:8]     # force to exactly 64 bits
        
        # encrypt
        s20 = Salsa20(key, IV, rounds)
        ciphertxt = s20.encryptBytes(message)
        
        # decrypt
        s20 = Salsa20(key, IV, rounds)
        plaintxt  = s20.encryptBytes(ciphertxt)
    
        if message == plaintxt:
            print '    *** good ***'
        else:
            print '    *** bad ***'


    if 1:               # one known good 8-round test vector 
        rounds = 8      # must be 8 for this test
        message = chr(0)*64
        key = '00000000000000000000000000000002'.decode('hex')
        IV  = '0000000000000000'.decode('hex')
        out64 = bytestring("""
               06C80B8CEC60F0C2E73EB6ED5DCB1B9C
               39B210F1AB76FEDF1A6B7AE370DA0F20
               0CEBCAD6EF6E57AC80E4375C035FA44D
               3AE4DC2C2507757DAF37B14F36643489""")
        s20 = Salsa20(key, IV, rounds)
        s20.setIV(IV)
        ciphertxt = s20.encryptBytes(message)
        s20 = Salsa20(key, IV, rounds)
        plaintxt  = s20.encryptBytes(ciphertxt)
        if (message == plaintxt and
            ciphertxt[:64] == out64):
            print '    *** vector 1 good ***'
        else:
            print '    *** vector 1 bad ***'
        
    if 1:               # one known good 8-round test vector 
        rounds = 8      # must be 8 for this test
        message = chr(0)*65536
        key = '0053A6F94C9FF24598EB3E91E4378ADD'.decode('hex')
        IV  = '0D74DB42A91077DE'.decode('hex')
        out64 = bytestring("""
               75FCAE3A3961BDC7D2513662C24ADECE
               995545599FF129006E7A6EE57B7F33A2
               6D1B27C51EA15E8F956693472DC23132
               FCD90FB0E352D26AF4DCE5427193CA26""")
        out65536 = bytestring("""
               EA75A566C431A10CED804CCD45172AD1
               EC4930E9869372B8EDDF303098A8910C
               EE123BF849C51A33554BA1445E6B6268
               4921F36B77EADC9681A2BB9DDFEC2FC8""")
        s20 = Salsa20(key, IV, rounds)
        ciphertxt = s20.encryptBytes(message)
        s20 = Salsa20(key, IV, rounds)
        plaintxt  = s20.encryptBytes(ciphertxt)
        if (message == plaintxt and
            ciphertxt[:64] == out64 and
            ciphertxt[65472:] == out65536):
            print '    *** vector 2 good ***'
        else:
            print '    *** vector 2 bad ***'
        
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

if __name__ == '__main__':
    test()

#--------------------------------------------------------------------------
#--------------------------------------------------------------------------
#--------------------------------------------------------------------------
